import { Context } from 'koa';
import CODE from '@utils/code';
import services from '../services';

export interface InterfaceDocument {
  method: number
  title: string
  url: string
  tag: string
  requestList?: InterfaceRequestList[]
  responseSuccess?: string
  responseError?: string
  _id?: string,
}
interface InterfaceRequestList {
  title: string
  type: string
  require: boolean
  defaultValue?: string
  content?: string
}

export interface InterfacePage {
  page: number
  size: number
}

export interface InterfaceTag {
  name: string
  _id?: string
}

export interface InterfaceDocumentList {
  _id: string
}

const putDocument = async (ctx: Context): Promise<void> => {
  const { method, title, url, requestList = [], tag }: InterfaceDocument = ctx.request.body;
  if (!method || !title || !url || !tag) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const errorRequestList = requestList.filter(el => (!el.title || !el.type));
  if (errorRequestList.length) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.putDocument(ctx, ctx.request.body);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const getDocument = async (ctx:Context): Promise<void> => {
  const { id } = ctx.params;
  if (!id) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '数据不存在', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.getDocument(ctx, id);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const getDocumentList = async (ctx:Context): Promise<void> => {
  const { _id } = ctx.query;
  const data = await services.document.getDocumentList(ctx, { _id: _id as string });
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const deleteDocument = async (ctx:Context): Promise<void> => {
  const { id } = ctx.params;
  if (!id) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '数据不存在', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.deleteDocument(ctx, id);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const updateDocument = async (ctx:Context): Promise<void> => {
  const { id: _id } = ctx.params;
  const { method, title, url, requestList = [] }: InterfaceDocument = ctx.request.body;
  if (!_id) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '数据不存在', false, CODE.MISS_PARAMS);
    return;
  }

  if (!method || !title || !url) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const errorRequestList = requestList.filter(el => (!el.title || !el.type));
  if (errorRequestList.length) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.updateDocument(ctx, { _id, ...ctx.request.body });
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const putTag = async (ctx:Context): Promise<void> => {
  const { name }: InterfaceTag = ctx.request.body;
  if (!name) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.putTag(ctx, ctx.request.body);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const updateTag = async (ctx:Context): Promise<void> => {
  const { id: _id } = ctx.params;
  const { name }: InterfaceTag = ctx.request.body;
  console.log(name, _id);
  if (!name || !_id) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '必填项参数缺失', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.updateTag(ctx, { _id, ...ctx.request.body });
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const getTagList = async (ctx:Context): Promise<void> => {
  const { page, size } = ctx.query;
  const params = {
    page: Number(page) || 1,
    size: Number(size) || 10
  };
  const data = await services.document.getTagList(ctx, params);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

const deleteTag = async (ctx:Context): Promise<void> => {
  const { id } = ctx.params;
  if (!id) {
    ctx.status = 400;
    ctx.body = ctx.$response(null, '数据不存在', false, CODE.MISS_PARAMS);
    return;
  }
  const data = await services.document.deleteTag(ctx, id);
  ctx.body = ctx.$response(data.data, data.message, data.success, data.code);
};

export default {
  putDocument,
  getDocument,
  getDocumentList,
  deleteDocument,
  updateDocument,
  putTag,
  updateTag,
  getTagList,
  deleteTag
};
