import log4js, { DateFileAppender, ConsoleAppender, Logger } from 'log4js';
import path from 'path';
import { Context } from 'koa';
import config from '@config';
export interface InterfaceLogInfo {
  ip: string,
  params?: unknown,
  time?: number
}
export interface InterfaceLogger {
  (ctx: Context): InterfaceLog
}
export interface InterfaceLog {
  trace: InterfaceLogFn
  debug: InterfaceLogFn
  info: InterfaceLogFn
  warn: InterfaceLogFn
  error: InterfaceLogFn
  fatal: InterfaceLogFn
  mark: InterfaceLogFn
}
interface InterfaceLogFn {
  (message: string, options?: { [key: string]: string|number }): void;
}
interface InterfaceAppenders {
  [key: string]: DateFileAppender | ConsoleAppender
}

// 输出格式化配置
export const formatOutputHandle = (
  ctx: Context,
  message: string | null | undefined ,
  commonInfo: InterfaceLogInfo
): string => {
  const {
    method, // 请求方式
    headers, // 请求headers
    origin, // 获取URL的来源，包括 protocol 和 host。
    originalUrl // 获取请求原始URL。
  } = ctx.request;
  const {
    time,
    ...newCommonInfo
   } = commonInfo;
  interface client {
    'content-type'?: string
    query?: string
  }
  const client: client = {};
  if (headers['content-type']) {
    client['content-type'] = headers['content-type'];
  }
  if (ctx.querystring) {
    client.query = ctx.querystring;
  }
  let text = `[${method}] ${ctx.status} [${origin}${originalUrl}] ${time ? time + 'ms' : ''}
req: ${JSON.stringify(Object.assign(client, newCommonInfo))}`;
  if ([
    '[object Object]',
    '[object Array]'
  ].includes(Object.prototype.toString.call(message))) {
    message = JSON.stringify(message);
  }
  if (Object.prototype.toString.call(message) === '[object Error]') {
    message = message?.toString();
  }
  if (message) {
    text += '\n' + String(message);
  }
  return text;
};

// 日志级别
// const methods: string[] = ['trace', 'debug', 'info', 'warn', 'error', 'fatal', 'mark'];
const methodsAcc: string[] = ['trace', 'debug', 'info', 'warn'];
const methodsErr: string[] = ['error', 'fatal', 'mark'];
// 默认配置
const defaultConfig = {
  logLevel: 'debug', // 日志级别
  dir: path.resolve(__dirname, '../log'), // 指定日志存放目录名
  env: 'development' // 指定当前环境，当开发时控制台也输出
};

let logger: Logger = {} as Logger;
let loggerErr: Logger = {} as Logger;

export function formatLog(
  type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  message: string ,
): void
export function formatLog(
  type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  message: string ,
  ctx?: Context,
  options?: { [key: string]: string|number }): void
export function formatLog(
  type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  message: string ,
  ctx?: Context,
  options?: { [key: string]: string|number }): void {
    let log: Logger;
    if (methodsErr.includes(type)) {
      log = loggerErr;
    } else {
      log = logger;
    }
    if (ctx) {
      log[type](formatOutputHandle(ctx, message, Object.assign({ ip: ctx.ip }, options)));
    } else {
      log[type](message);
    }
}

export default class MyLogger {
  // log配置
  logConfig = Object.assign({}, defaultConfig, config.logs);
  // 记录日志的方式
  appenders: InterfaceAppenders = {}
  // 记录日志的类型
  categories
  // 是否是开发环境
  isDev = false
  // log4js实例
  logger: Logger = {} as Logger
  loggerErr: Logger = {} as Logger
  constructor() {
    this.init();
  }
  init(): void {
    const { env, appLogLevel, dir } = this.logConfig;
    /**
     * 记录日志的方式
     * 指定要记录的日志分类 log
     * 展示方式为文件类型 dateFile
     * 日志输出的文件名 s-yyyy-MM-dd.log
     */
    // interface appenders {
    //   [key: string]: {
    //     type: string,
    //     filename?: string,
    //     pattern?: string,
    //     alwaysIncludePattern?: boolean,
    //     ip?: string,
    //     layout: object
    //   }
    // }
    const appenders: InterfaceAppenders = {
      access: {
        // 按日期进行输出log
        type: 'dateFile',
        // 路径和文件名
        filename: `${dir}/access/z-acc`,
        pattern: 'yyyy-MM-dd.log',
        alwaysIncludePattern: true,
        // 日志输出格式
        layout: {
          type: 'pattern',
          pattern: '[%d{yyyy-MM-dd hh:mm:ss}] [%p] %c %m%n'
        }
      },
      error: {
        // 按日期进行输出log
        type: 'dateFile',
        // 路径和文件名
        filename: `${dir}/error/z-err`,
        pattern: 'yyyy-MM-dd.log',
        alwaysIncludePattern: true,
        // 日志输出格式
        layout: {
          type: 'pattern',
          pattern: '[%d{yyyy-MM-dd hh:mm:ss}] [%p] %c %m%n'
        }
      }
    };
    // 开发和本地同时启用console
    const isDev = env === 'dev' || env === 'local' || env === 'development';
    if (isDev) {
      appenders.out = {
        type: 'console',
        layout: {
          type: 'pattern',
          pattern: '%[[%d{yyyy-MM-dd hh:mm:ss}] [%p] %c%] %m%n'
        }
      };
    }
    this.isDev = isDev;
    this.appenders = appenders;
    /**
     * 指定日志的默认配置项
     * 如果 log4js.getLogger 中没有指定，默认为 appenders中的日志配置项, 数组
     * 指定日志的记录内容显示 某级别 及 某级别 以上级别的信息
     */
    const categories = {
      default: {
        // appenders: Object.keys(appenders),
        appenders: ['access'],
        level: appLogLevel
      },
      error: {
        appenders: ['error'],
        level: appLogLevel
      }
    };
    this.categories = categories;
    // log4js配置
    log4js.configure({
      appenders,
      categories,
      pm2: !isDev,
      pm2InstanceVar: 'INSTANCE_ID'
    });
    // 初始化log4js
    logger = log4js.getLogger();
    loggerErr = log4js.getLogger('error');
    this.logger = logger;
    this.loggerErr = loggerErr;
  }
  middleware(ctx: Context): InterfaceLog {
    const $log: InterfaceLog = {
      trace: () => {},
      debug: () => {},
      info: () => {},
      warn: () => {},
      error: () => {},
      fatal: () => {},
      mark: () => {}
    };
    methodsAcc.map(el => {
      $log[el] = (message, options?: { [key: string]: string }) => {
        this.logger[el](formatOutputHandle(ctx, message, Object.assign({ ip: ctx.ip }, options)));
      };
    });
    methodsErr.map(el => {
      $log[el] = (message, options?: { [key: string]: string }) => {
        this.loggerErr[el](formatOutputHandle(
          ctx,
          message,
          Object.assign({ ip: ctx.ip },
          options
        )));
      };
    });
    return $log;
  }
  formatLog
  // formatLog(
  //   this: MyLogger,
  //   type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  //   message: string ,
  // ): void
  // formatLog(
  //   this: MyLogger,
  //   type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  //   message: string ,
  //   ctx?: Context,
  //   options?: { [key: string]: string|number }): void
  // formatLog(
  //   this: MyLogger,
  //   type: 'trace'|'debug'|'info'|'warn'|'error'|'fatal'|'mark',
  //   message: string ,
  //   ctx?: Context,
  //   options?: { [key: string]: string|number }): void {
  //     let log: Logger;
  //     if (methodsErr.includes(type)) {
  //       log = loggerErr;
  //     } else {
  //       log = logger;
  //     }
  //     if (ctx) {
  //       log[type](formatOutputHandle(ctx, message, Object.assign({ ip: ctx.ip }, options)));
  //     } else {
  //       log[type](message);
  //     }
  // }
}
