import path from 'path';
// 将多个流拼接成一个流
import MultiStream from 'multistream';
import {
  closeSync,
  createReadStream,
  createWriteStream,
  ensureDir,
  existsSync,
  mkdir,
  openSync,
  readdir,
  remove,
  rename,
  writeFile
} from 'fs-extra';
import { calculateMd5, calculateFileMd5 } from '@utils/cipher';
import config from '@config';

export enum FILE_UPLOADER_STATUS_CODE {
  SUCCESS = 0,
  FOLDER_EXIST = 1001,
  NOT_FOUND = 1002,
  MD5_CHECK_FAILED = 1003,
  UNKNOWN_ERROR = 5000,
}

export interface IFileInfo {
  name: string;
  path: string;
}

export interface IFileUploaderOptions {
  tempFileLocation: string;
  mergedFileLocation: string;
}

export interface IUploadPartInfo {
  path: string;
  index: number;
  md5: string;
}

export interface IUploadPartInfoReturn {
  partFileMd5: string,
  partIndex: number
}

export interface IMergedFileInfo {
  path: string;
  md5: string;
}

export class Exception extends Error {
  code: FILE_UPLOADER_STATUS_CODE;
  constructor(code: FILE_UPLOADER_STATUS_CODE, message: string) {
    super(message);
    this.code = code;
    this.message = message;
  }
}

export class FolderExistException extends Exception {
  constructor(message: string) {
    super(FILE_UPLOADER_STATUS_CODE.FOLDER_EXIST, message);
  }
}

export class NotFoundException extends Exception {
  constructor(message: string) {
    super(FILE_UPLOADER_STATUS_CODE.NOT_FOUND, message);
  }
}

export class Md5Exception extends Exception {
  constructor(message: string) {
    super(FILE_UPLOADER_STATUS_CODE.MD5_CHECK_FAILED, message);
  }
}

function wait(time: number): Promise<void> {
  return new Promise<void>((resolve) => {
    setTimeout(() => {
      resolve();
    }, time);
  });
}

async function listDir(path: string): Promise<IFileInfo[]> {
  const items = await readdir(path);
  return Promise.all(
    items
      .filter((item: string) => !item.startsWith('.'))
      .map(async (item: string) => {
        return {
          name: item,
          path: `${path}/${item}`,
        };
      }),
  );
}

async function mergePartFile(
  files: IFileInfo[],
  mergedFilePath: string,
): Promise<void> {
  const fileList = files.map((item) => {
    const [index] = item.name.replace(/\.part$/, '').split('|');
    return {
      index: parseInt(index),
      path: item.path,
    };
  });
  const sortedFileList = fileList.sort((a, b) => {
    return a.index - b.index;
  });
  const sortedFilePathList = sortedFileList.map((item) => item.path);
  merge(sortedFilePathList, mergedFilePath);
}

function merge(inputPathList: string[], outputPath: string) {
  const fd = openSync(outputPath, 'w+');
  const writeStream = createWriteStream(outputPath);
  const readStreamList = inputPathList.map((path) => {
    return createReadStream(path);
  });
  return new Promise((resolve, reject) => {
    const multiStream: NodeJS.ReadableStream = new (MultiStream as any)(readStreamList);
    multiStream.pipe(writeStream);
    multiStream.on('end', () => {
      closeSync(fd);
      resolve(true);
    });
    multiStream.on('error', () => {
      closeSync(fd);
      reject(false);
    });
  });
}


export class FileUploaderServer {
  private fileUploaderOptions: IFileUploaderOptions = {
    tempFileLocation: config.bigFileUpload.defaultTempFileLocation
      .replace('$suffixDir', config.bigFileUpload.suffixDir),
    mergedFileLocation: config.bigFileUpload.defaultMergedFileLocation
      .replace('$suffixDir', config.bigFileUpload.suffixDir),
  };
  constructor(options?: IFileUploaderOptions) {
    this.fileUploaderOptions = Object.assign(this.fileUploaderOptions, options);
  }

  /**
   * 获取配置项
   */
  public getOptions() :IFileUploaderOptions {
    return this.fileUploaderOptions;
  }

  /**
   * 初始化文件分片上传，实际上就是根据fileName和时间计算一个md5，并新建一个文件夹
   * @param fileName 文件名
   * @returns 上传Id
   */
  public async initFilePartUpload(fileName: string): Promise<string> {
    const { tempFileLocation } = this.fileUploaderOptions;
    await ensureDir(tempFileLocation);
    const uploadId = calculateMd5(`${fileName}-${Date.now()}`);
    const uploadFolderPath = path.join(tempFileLocation, uploadId);
    const uploadFolderExist = existsSync(uploadFolderPath);
    if (uploadFolderExist) {
      throw new FolderExistException(
        'found same upload folder, maybe you meet hash collision',
      );
    }
    await mkdir(uploadFolderPath);
    return uploadId;
  }

  /**
   * 上传分片，实际上是将partFile写入uploadId对应的文件夹中，写入的文件命名格式为`partIndex|md5`
   * @param uploadId 上传Id
   * @param partIndex 分片序号
   * @param partFile 分片内容
   * @returns 分片md5
   */
  public async uploadPartFile(
    uploadId: string,
    partIndex: number,
    partFile: Buffer,
  ): Promise<IUploadPartInfoReturn> {
    const uploadFolderPath = await this.getUploadFolder(uploadId);
    const partFileMd5 = calculateMd5(partFile);
    const partFileLocation = path.join(
      uploadFolderPath,
      `${partIndex}|${partFileMd5}.part`,
    );
    await writeFile(partFileLocation, partFile);
    return {
      partFileMd5,
      partIndex
    };
  }

  /**
   * 获取已上传的分片信息，实际上就是读取这个文件夹下面的内容
   * @param uploadId 上传Id
   * @returns 已上传的分片信息
   */
  public async listUploadedPartFile(
    uploadId: string,
  ): Promise<IUploadPartInfo[]> {
    const uploadFolderPath = await this.getUploadFolder(uploadId);
    const dirList = await listDir(uploadFolderPath);
    const uploadPartInfo = dirList.map((item: IFileInfo) => {
      const [index, md5] = item.name.replace(/\.part$/, '').split('|');
      return {
        path: item.path,
        index: parseInt(index),
        md5,
      };
    });
    return uploadPartInfo;
  }

  /**
   * 取消文件上传，硬删除会直接删除文件夹，软删除会给文件夹改个名字
   * @param uploadId 上传Id
   * @param deleteFolder 是否直接删除文件夹
   */
  async cancelFilePartUpload(
    uploadId: string,
    deleteFolder = false,
  ): Promise<void> {
    const uploadFolderPath = await this.getUploadFolder(uploadId);
    if (deleteFolder) {
      await remove(uploadFolderPath);
    } else {
      await rename(uploadFolderPath, `${uploadFolderPath}[removed]`);
    }
  }

  /**
   * 完成分片上传，实际上就是将所有分片都读到一起，然后进行md5检查，最后存到一个新的路径下。
   * @param uploadId 上传Id
   * @param fileName 文件名
   * @param md5 文件md5
   * @returns 文件存储路径
   */
  async finishFilePartUpload(
    uploadId: string,
    fileName: string,
    md5: string,
  ): Promise<IMergedFileInfo> {
    const { mergedFileLocation } = this.fileUploaderOptions;
    await ensureDir(mergedFileLocation);
    const uploadFolderPath = await this.getUploadFolder(uploadId);
    const dirList = await listDir(uploadFolderPath);
    const files = dirList.filter((item) => item.path.endsWith('.part'));
    const mergedFileDirLocation = path.join(mergedFileLocation, md5);
    await ensureDir(mergedFileDirLocation);
    const mergedFilePath = path.join(mergedFileDirLocation, fileName);
    await mergePartFile(files, mergedFilePath);
    await wait(1000); // 要等待一段时间，否则在计算md5时会读取到空文件
    const mergedFileMd5 = await calculateFileMd5(mergedFilePath);
    if (mergedFileMd5 !== md5) {
      throw new Md5Exception('md5 checked failed');
    }
    return {
      path: mergedFilePath,
      md5,
    };
  }

  /**
   * 获取上传文件夹的路径
   * @param uploadId 上传Id
   * @returns 文件夹路径
   */
  private async getUploadFolder(uploadId: string): Promise<string> {
    if (!uploadId || uploadId === 'undefined') {
      throw new NotFoundException('uploadId 不存在');
    }
    const { tempFileLocation } = this.fileUploaderOptions;
    await ensureDir(tempFileLocation);
    const uploadFolderPath = path.join(tempFileLocation, uploadId);
    const uploadFolderExist = existsSync(uploadFolderPath);
    if (!uploadFolderExist) {
      throw new NotFoundException('not found upload folder');
    }
    return uploadFolderPath;
  }
}
